// Copyright 2019 Robert Bosch GmbH
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/** \mainpage tracetools: tracing tools and instrumentation for ROS 2
 *
 * `tracetools` provides utilities to instrument ROS.
 * It provides two main headers:
 *
 * - tracetools/tracetools.h
 *   - instrumentation functions
 * - tracetools/utils.hpp
 *   - utility functions
 */

#ifndef TRACETOOLS__TRACETOOLS_H_
#define TRACETOOLS__TRACETOOLS_H_

#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include "tracetools/config.h"
#include "tracetools/visibility_control.hpp"

#ifndef TRACETOOLS_DISABLED
/// Call a tracepoint.
/**
 * This is the preferred method over calling the actual function directly.
 */
#  define TRACEPOINT(event_name, ...) \
  (ros_trace_ ## event_name)(__VA_ARGS__)
#  define TRACEPOINT_NOARGS(event_name) \
  (ros_trace_ ## event_name)()
#  define DECLARE_TRACEPOINT(event_name, ...) \
  TRACETOOLS_PUBLIC void ros_trace_ ## event_name(__VA_ARGS__); \
  TRACETOOLS_PUBLIC bool ros_trace_enabled_ ## event_name(); \
  TRACETOOLS_PUBLIC void ros_trace_do_ ## event_name(__VA_ARGS__);
#  define DECLARE_TRACEPOINT_NOARGS(event_name) \
  TRACETOOLS_PUBLIC void ros_trace_ ## event_name(void); \
  TRACETOOLS_PUBLIC bool ros_trace_enabled_ ## event_name(); \
  TRACETOOLS_PUBLIC void ros_trace_do_ ## event_name(void);

#else
#  define TRACEPOINT(event_name, ...) ((void) (0))
#  define DECLARE_TRACEPOINT(event_name, ...)
#endif

/// Check if a tracepoint is enabled at runtime.
/**
 * This can be useful to only compute tracepoint arguments if the tracepoint is actually enabled at
 * runtime. Combine this with `DO_TRACEPOINT()` instead of `TRACEPOINT()` to trigger a tracepoint
 * after checking if it is enabled and computing its arguments.
 *
 * This is the preferred method over calling the underlying function directly.
 */
#  define TRACEPOINT_ENABLED(event_name) \
  ros_trace_enabled_ ## event_name()

#ifdef __cplusplus
extern "C"
{
#endif

/// Get tracing compilation status.
/**
 * \return `true` if tracing is enabled, `false` otherwise
 */
TRACETOOLS_PUBLIC bool ros_trace_compile_status();

/// `rcl_init`
/**
 * Initialisation for the whole process.
 * Notes the `tracetools` version automatically.
 *
 * \param[in] context_handle pointer to the `rcl_context_t` handle
 */
DECLARE_TRACEPOINT(
  rcl_init,
  const void * context_handle)

/// `rcl_node_init`
/**
 * Node initialisation.
 * Links a `rcl_node_t` handle to its `rmw_node_t` handle.
 *
 * \param[in] node_handle pointer to the node's `rcl_node_t` handle
 * \param[in] rmw_handle pointer to the node's `rmw_node_t` handle
 * \param[in] node_name node name
 * \param[in] node_namespace node namespace
 */
DECLARE_TRACEPOINT(
  rcl_node_init,
  const void * node_handle,
  const void * rmw_handle,
  const char * node_name,
  const char * node_namespace)

/// `rcl_publisher_init`
/**
 * Publisher initialisation.
 * Links a `rcl_publisher_t` handle to its `rcl_node_t` handle
 * and its `rmw_publisher_t` handle, and links it to a topic name.
 *
 * \param[in] publisher_handle pointer to the publisher's `rcl_publisher_t` handle
 * \param[in] node_handle pointer to the `rcl_node_t` handle of the node the publisher belongs to
 * \param[in] rmw_publisher_handle pointer to the publisher's `rmw_publisher_t` handle
 * \param[in] topic_name full topic name
 * \param[in] queue_depth publisher history depth
 */
DECLARE_TRACEPOINT(
  rcl_publisher_init,
  const void * publisher_handle,
  const void * node_handle,
  const void * rmw_publisher_handle,
  const char * topic_name,
  const size_t queue_depth)

/// `rcl_publish`
/**
 * Message publication.
 * Links a `rcl_publisher_t` handle to a pointer to
 * a message being published at the `rcl` level.
 *
 * \param[in] publisher_handle pointer to the publisher's `rcl_publisher_t` handle
 * \param[in] message pointer to the message being published
 */
DECLARE_TRACEPOINT(
  rcl_publish,
  const void * publisher_handle,
  const void * message)

/// `rcl_init`
/**
 * Initialisation for the whole process.
 * Notes the `tracetools` version automatically.
 *
 * \param[in] context_handle pointer to the `rcl_context_t` handle
 */
DECLARE_TRACEPOINT(
  rcl_init,
  const void * context_handle)

/// `rclcpp_publish`
/**
 * Message publication.
 * Links a `rcl_publisher_t` handle to a pointer to
 * a message being published at the `rclcpp` level.
 *
 * \param[in] publisher_handle pointer to the publisher's `rcl_publisher_t` handle
 * \param[in] message pointer to the message being published
 */

DECLARE_TRACEPOINT(
  rclcpp_publish,
  const void * publisher_handle,
  const void * message)

/// `rclcpp_executor_execute`
/**
 * Executable execution.
 * Notes an executable being executed using its `rcl` handle, which can be a:
 *   * timer
 *   * subscription
 *
 * \param[in] handle pointer to the `rcl` handle of the executable being executed
 */
DECLARE_TRACEPOINT(
  rclcpp_executor_execute,
  const void * handle)

/// `rclcpp_executor_wait_for_work`
/**
 * Notes the start time of the executor phase that waits for work and notes the timeout value.
 *
 * \param[in] timeout the timeout value for the wait call
 */
DECLARE_TRACEPOINT(
  rclcpp_executor_wait_for_work,
  const int64_t timeout)

/// `rclcpp_take`
/**
 * Message taking.
 * Notes the pointer to the message being taken at the `rclcpp` level.
 *
 * \param[in] message pointer to the message being taken
 */
DECLARE_TRACEPOINT(
  rclcpp_take,
  const void * message)

/// `rclcpp_executor_get_next_ready`
/**
 * Notes the start time of the executor phase that gets the next executable that's ready.
 */
DECLARE_TRACEPOINT_NOARGS(
  rclcpp_executor_get_next_ready)

/// `rmw_take`
/**
 * Message taking.
 * Links a `rmw_subscription_t` handle to a pointer to a message being taken at the `rmw` level.
 * Notes the source timestamp of the message.
 *
 * \param[in] rmw_subscription_handle pointer to the subscription's `rmw_subscription_t` handle
 * \param[in] message pointer to the message being taken
 * \param[in] source_timestamp the source timestamp of the message,
 *  or 0 (if no message or no info)
 * \param[in] taken whether a message was taken
 */
DECLARE_TRACEPOINT(
  rmw_take,
  const void * rmw_subscription_handle,
  const void * message,
  int64_t source_timestamp,
  const bool taken)

/// `rmw_publish`
/**
 * Message publication.
 * Notes the pointer to the message being published at the `rmw` level.
 * Also notes the source timestamp of the message.
 *
 * \param[in] rmw_publisher_handle pointer to the publisher's `rmw_publisher_t` handle
 * \param[in] message pointer to the message being published
 * \param[in] timestamp the source timestamp of the message
 */
DECLARE_TRACEPOINT(
  rmw_publish,
  const void * rmw_publisher_handle,
  const void * message,
  int64_t timestamp)

/// `rmw_publisher_init`
/**
 * RMW publisher initialisation.
 * Links a `rmw_publisher_t` handle to its DDS/rmw GID.
 *
 * \param[in] rmw_publisher_handle pointer to the publisher's `rmw_publisher_t` handle
 * \param[in] gid pointer to the publisher's DDS/rmw GID
 */
DECLARE_TRACEPOINT(
  rmw_publisher_init,
  const void * rmw_publisher_handle,
  const uint8_t * gid)

/// `rmw_client_init`
/**
 * RMW client initialisation.
 * Links a `rmw_client_t` handle to its DDS/rmw GID.
 * This GID should be the same client GID that is collected by the `rmw_take_request` and
 * `rmw_send_response` tracepoints for requests made by this client and responses to requests made
 * by this client.
 *
 * \param[in] rmw_client_handle pointer to the client's `rmw_client_t` handle
 * \param[in] gid pointer to the client's DDS/rmw GID
 */
DECLARE_TRACEPOINT(
  rmw_client_init,
  const void * rmw_client_handle,
  const uint8_t * gid)

/// `rmw_subscription_init`
/**
 * RMW subscription initialisation.
 * Links a `rmw_subscription_handle` handle to its DDS/rmw GID.
 *
 * \param[in] rmw_subscription_handle pointer to the publisher's `rmw_subscription_t` handle
 * \param[in] gid pointer to the subscription's DDS/rmw GID
 */
DECLARE_TRACEPOINT(
  rmw_subscription_init,
  const void * rmw_subscription_handle,
  const uint8_t * gid)

/// `rmw_take_response`
/**
 * Response taking.
 * Links a `rmw_client_t` handle to a pointer to a response being taken at the `rmw` level.
 * Notes the source timestamp of the response. Also notes the sequence number of the request this
 * response is for. It does not note the request's client GID, since it is assumed that the matching
 * of the response to the original client is performed before this tracepoint.
 *
 * \param[in] rmw_client_handle pointer to the client's `rmw_client_t` handle
 * \param[in] response pointer to the response being taken
 * \param[in] sequence_number the sequence number of the request this response is for,
 *  or 0 (if no response or no info)
 * \param[in] source_timestamp the source timestamp of the response,
 *  or 0 (if no response or no info)
 * \param[in] taken whether a response was taken
 */
DECLARE_TRACEPOINT(
  rmw_take_response,
  const void * rmw_client_handle,
  const void * response,
  int64_t sequence_number,
  int64_t source_timestamp,
  const bool taken)

/// `rmw_take_request`
/**
 * Request taking.
 * Links a `rmw_service_t` handle to a pointer to a request being taken at the `rmw` level.
 * Also notes the sequence number of the request and the GID of the requesting client.
 *
 * \param[in] rmw_service_handle pointer to the service's `rmw_service_t` handle
 * \param[in] request pointer to the request being taken
 * \param[in] client_gid the GID of the requesting client
 * \param[in] sequence_number the sequence number of the received request,
 *  or 0 (if no request or no info)
 * \param[in] taken whether a request was taken
 */
DECLARE_TRACEPOINT(
  rmw_take_request,
  const void * rmw_service_handle,
  const void * request,
  const uint8_t * client_gid,
  int64_t sequence_number,
  const bool taken)

/// `rmw_send_response`
/**
 * Response publication.
 * Notes the pointer to the response being sent at the `rmw` level.
 * Also notes the sequence number of the request that this response is for and the GID of the
 * requesting client, and notes the source timestamp of the response.
 *
 * \param[in] rmw_service_handle pointer to the service's `rmw_service_t` handle
 * \param[in] response pointer to the request being sent
 * \param[in] client_gid the GID of the requesting client
 * \param[in] sequence_number the sequence number of the request this response is for
 * \param[in] timestamp the source timestamp of the response
 */
DECLARE_TRACEPOINT(
  rmw_send_response,
  const void * rmw_service_handle,
  const void * response,
  const uint8_t * client_gid,
  int64_t sequence_number,
  int64_t timestamp)

/// `rmw_send_request`
/**
 * Request publication.
 * Notes the pointer to the request being sent at the `rmw` level.
 * Also notes the sequence number of the request.
 *
 * \param[in] rmw_client_handle pointer to the client's `rmw_client_t` handle
 * \param[in] request pointer to the request being sent
 * \param[in] sequence_number the sequence number of the request
 */
DECLARE_TRACEPOINT(
  rmw_send_request,
  const void * rmw_client_handle,
  const void * request,
  int64_t sequence_number)

/// `rcl_take`
/**
 * Message taking.
 * Notes the pointer to the message being taken at the `rcl` level.
 *
 * \param[in] message pointer to the message being taken
 */
DECLARE_TRACEPOINT(
  rcl_take,
  const void * message)

/// `rclcpp_intra_publish`
/**
 * Intra-process message publication.
 * Notes the pointer to the message being published in intra process at the `rclcpp` level.
 *
 * \param[in] publisher_handle pointer to the publisher's `rcl_publisher_t` handle
 * \param[in] message pointer to the message being published
 */
DECLARE_TRACEPOINT(
  rclcpp_intra_publish,
  const void * publisher_handle,
  const void * message)

/// `rclcpp_ring_buffer_clear`
/**
 * Ring buffer clear.
 * Notes the address of the cleared buffer.
 *
 * \param[in] buffer pointer to the buffer
 */
DECLARE_TRACEPOINT(
  rclcpp_ring_buffer_clear,
  const void * buffer)

/// `message_link_periodic_async`
/**
 * Message link annotation for N-to-M link.
 * 
 * \param[in] subs pointers to the `rcl` handles of the subscriptions in the link
 * \param[in] num_subs the number of subscriptions in the link
 * \param[in] pubs pointers to the `rcl` handles of the publishers in the link
 * \param[in] num_pubs the number of publishers in the link
 */
DECLARE_TRACEPOINT(
  message_link_periodic_async,
  const void ** subs,
  const size_t num_subs,
  const void ** pubs,
  const size_t num_pubs)

/// `message_link_partial_sync`
/**
 * Message link annotation for 2-to-M link.
 * 
 * \param[in] subs pointers to the `rcl` handles of the subscriptions in the link
 * \param[in] num_subs the number of subscriptions in the link
 * \param[in] pubs pointers to the `rcl` handles of the publishers in the link
 * \param[in] num_pubs the number of publishers in the link
 */
DECLARE_TRACEPOINT(
  message_link_partial_sync,
  const void ** subs,
  const size_t num_subs,
  const void ** pubs,
  const size_t num_pubs)

/// `rclcpp_ring_buffer_dequeue`
/**
  * Ring buffer dequeue.
  * Notes buffer address, the index to read from, and the size of the buffer after this operation.
  *
  * \param[in] buffer pointer to the buffer
  * \param[in] index the index to read from
  * \param[in] size the size of the buffer after this operation
  */
DECLARE_TRACEPOINT(
  rclcpp_ring_buffer_dequeue,
  const void * buffer,
  const uint64_t index,
  const uint64_t size)

/// `rclcpp_buffer_to_ipb`
/**
 * Intra-process buffer initialization.
 * Notes the `BufferImplementationBase` and the `IntraProcessBuffer` that owns it.
 *
 * \param[in] buffer the pointer to the `BufferImplementationBase`
 * \param[in] ipb to pointer to the `IntraProcessBuffer`
 */
DECLARE_TRACEPOINT(
  rclcpp_buffer_to_ipb,
  const void * buffer,
  const void * ipb)

/// `rclcpp_construct_ring_buffer`
/**
 * Ring buffer construction.
 * Notes the buffer address and its capacity.
 *
 * \param[in] buffer pointer to the buffer
 * \param[in] capacity buffer size
 */
DECLARE_TRACEPOINT(
  rclcpp_construct_ring_buffer,
  const void * buffer,
  const uint64_t capacity)

/// `rclcpp_ring_buffer_enqueue`
/**
 * Notes buffer address, the index to write to, and the occurrence of the lost.
 *
 * \param[in] buffer pointer to the buffer
 * \param[in] index the index to write to
 * \param[in] size the size of the buffer after this operation
 * \param[in] overwritten occurrence of the lost
 */
DECLARE_TRACEPOINT(
  rclcpp_ring_buffer_enqueue,
  const void * buffer,
  const uint64_t index,
  const uint64_t size,
  const bool overwritten)

/// `rclcpp_timer_link_node`
/**
 * Link a timer to a node.
 *
 * \param[in] timer_handle pointer to the timer's `rcl_timer_t` handle
 * \param[in] node_handle pointer to the `rcl_node_t` handle of the node the timer belongs to
 */
DECLARE_TRACEPOINT(
  rclcpp_timer_link_node,
  const void * timer_handle,
  const void * node_handle)

/// `rcl_lifecycle_transition`
/**
 * Lifecycle transition between two states.
 *
 * \param[in] state_machine pointer to the state machine for the transition
 * \param[in] start_label start state label
 * \param[in] goal_label goal state label
 */
DECLARE_TRACEPOINT(
  rcl_lifecycle_transition,
  const void * state_machine,
  const char * start_label,
  const char * goal_label)

/// `rcl_lifecycle_state_machine_init`
/**
 * Lifecycle state machine initialisation.
 * Links a `rcl_lifecycle_state_machine_t` handle to a `rcl_node_t` handle.
 *
 * \param[in] node_handle pointer to the node handle
 * \param[in] state_machine pointer to the state machine
 */
DECLARE_TRACEPOINT(
  rcl_lifecycle_state_machine_init,
  const void * node_handle,
  const void * state_machine)

/// `rcl_subscription_init`
/**
 * Subscription initialisation.
 * Links a `rcl_subscription_t` handle to its `rcl_node_t` handle
 * and its `rmw_subscription_t` handle, and links it to a topic name.
 *
 * \param[in] subscription_handle pointer to the subscription's `rcl_subscription_t` handle
 * \param[in] node_handle
 *  pointer to the `rcl_node_t` handle of the node the subscription belongs to
 * \param[in] rmw_subscription_handle pointer to the subscription's `rmw_subscription_t` handle
 * \param[in] topic_name full topic name
 * \param[in] queue_depth subscription history depth
 */
DECLARE_TRACEPOINT(
  rcl_subscription_init,
  const void * subscription_handle,
  const void * node_handle,
  const void * rmw_subscription_handle,
  const char * topic_name,
  const size_t queue_depth)

/// `rclcpp_subscription_init`
/**
 * Subscription object initialisation.
 * Links the `rclcpp::*Subscription*` object to a `rcl_subscription_t` handle.
 * Needed since there could be more than 1 `rclcpp::*Subscription*` object
 * for one `rcl` subscription (e.g. when using intra-process).
 *
 * \param[in] subscription_handle
 *  pointer to the `rcl_subscription_t` handle of the subscription this object belongs to
 * \param[in] subscription pointer to this subscription object (e.g. `rclcpp::*Subscription*`)
 */
DECLARE_TRACEPOINT(
  rclcpp_subscription_init,
  const void * subscription_handle,
  const void * subscription)

/// `rclcpp_subscription_callback_added`
/**
 * Link a subscription callback object to a subscription object.
 *
 * \param[in] subscription pointer to the subscription object this callback belongs to
 * \param[in] callback pointer to this callback object (e.g. `rclcpp::AnySubscriptionCallback`)
 */
DECLARE_TRACEPOINT(
  rclcpp_subscription_callback_added,
  const void * subscription,
  const void * callback)

/// `rcl_service_init`
/**
 * Service initialisation.
 * Links a `rcl_service_t` handle to its `rcl_node_t` handle
 * and its `rmw_service_t` handle, and links it to a service name.
 *
 * \param[in] service_handle pointer to the service's `rcl_service_t` handle
 * \param[in] node_handle pointer to the `rcl_node_t` handle of the node the service belongs to
 * \param[in] rmw_service_handle pointer to the service's `rmw_service_t` handle
 * \param[in] service_name full service name
 */
DECLARE_TRACEPOINT(
  rcl_service_init,
  const void * service_handle,
  const void * node_handle,
  const void * rmw_service_handle,
  const char * service_name)

/// `rclcpp_service_callback_added`
/**
 * Link a service callback object to a service.
 *
 * \param[in] service_handle
 *  pointer to the `rcl_service_t` handle of the service this callback belongs to
 * \param[in] callback pointer to this callback object (e.g. `rclcpp::AnyServiceCallback`)
 */
DECLARE_TRACEPOINT(
  rclcpp_service_callback_added,
  const void * service_handle,
  const void * callback)

/// `rcl_client_init`
/**
 * Client initialisation.
 * Links a `rcl_client_t` handle to its `rcl_node_t` handle
 * and its `rmw_client_t` handle, and links it to a client name.
 *
 * \param[in] client_handle pointer to the client's `rcl_client_t` handle
 * \param[in] node_handle pointer to the `rcl_node_t` handle of the node the client belongs to
 * \param[in] rmw_client_handle pointer to the client's `rmw_client_t` handle
 * \param[in] service_name full client name
 */
DECLARE_TRACEPOINT(
  rcl_client_init,
  const void * client_handle,
  const void * node_handle,
  const void * rmw_client_handle,
  const char * service_name)

/// `rcl_timer_init`
/**
 * Timer initialisation.
 * Notes the timer's period.
 *
 * \param[in] timer_handle pointer to the timer's `rcl_timer_t` handle
 * \param[in] period period in nanoseconds
 */
DECLARE_TRACEPOINT(
  rcl_timer_init,
  const void * timer_handle,
  int64_t period)

/// `rclcpp_timer_callback_added`
/**
 * Link a timer callback object to its `rcl_timer_t` handle.
 *
 * \param[in] timer_handle
 *  pointer to the `rcl_timer_t` handle of the timer this callback belongs to
 * \param[in] callback pointer to the callback object (`std::function`)
 */
DECLARE_TRACEPOINT(
  rclcpp_timer_callback_added,
  const void * timer_handle,
  const void * callback)

/// `rclcpp_callback_register`
/**
 * Register a demangled function symbol with a callback.
 *
 * \param[in] callback pointer to the callback object
 *  (e.g. `rclcpp::AnySubscriptionCallback`,
 *  `rclcpp::AnyServiceCallback`, timer `std::function`, etc.)
 * \param[in] function_symbol demangled symbol of the callback function/lambda,
 *  see \ref get_symbol()
 */
DECLARE_TRACEPOINT(
  rclcpp_callback_register,
  const void * callback,
  const char * function_symbol)

/// `callback_start`
/**
 * Start of a callback.
 *
 * \param[in] callback pointer to this callback object
 *  (e.g. `rclcpp::AnySubscriptionCallback`,
 *  `rclcpp::AnyServiceCallback`, timer `std::function`, etc.)
 * \param[in] is_intra_process whether this callback is done via intra-process or not
 */
DECLARE_TRACEPOINT(
  callback_start,
  const void * callback,
  const bool is_intra_process)

/// `callback_end`
/**
 * End of a callback.
 *
 * \param[in] callback pointer to this callback object
 *  (e.g. `rclcpp::AnySubscriptionCallback`,
 *  `rclcpp::AnyServiceCallback`, timer `std::function`, etc.)
 */
DECLARE_TRACEPOINT(
  callback_end,
  const void * callback)

#ifdef __cplusplus
}
#endif

#endif  // TRACETOOLS__TRACETOOLS_H_
